@@ -4,17 +4,17 @@ import time
 import json
 import operator
 import urllib
+import csv
 
 from twisted.internet import threads, defer
 from twisted.web import resource, server
 from twisted.web.resource import NoResource
 import pgdb as sql
 
-import readingdb as rdb
-
 import smap.util as util
 import data
 import queryparse as qp
+import settings
 
 def makeErrback(request_):
     request = request_
@@ -99,21 +99,21 @@ class DataRequester:
                             qlimit = 1000000
 
                         while True:
-                            data = rdb.db_query(db, streamid_, qstart / 1000, end_ / 1000)
+                            data = settings.rdb.db_query(db, streamid_, qstart / 1000, end_ / 1000)
                             rv += data[:min(len(data), qlimit)]
                             qlimit -= min(len(data), qlimit)
                             if len(data) < 10000 or \
                                qlimit <= 0: break
-                            qstart = (rv[-1][0])*1000
+                            qstart = (rv[-1][0] + 1) * 1000
                         return request, rv
                     except:
                         traceback.print_exc()
                 elif method == 'next':
                     if qlimit == -1: qlimit = 1
-                    return request, rdb.db_next(db, streamid_, start_ / 1000, n = qlimit)
+                    return request, settings.rdb.db_next(db, streamid_, start_ / 1000, n = qlimit)
                 elif method == 'prev':
                     if qlimit == -1: qlimit = 1
-                    return request, rdb.db_prev(db, streamid_, start_ / 1000, n = qlimit)
+                    return request, settings.rdb.db_prev(db, streamid_, start_ / 1000, n = qlimit)
                 return request, []
             return queryFunc
 
@@ -279,9 +279,14 @@ class Api(resource.Resource):
         resource.Resource.__init__(self)
 
     def generic_extract_result(self, request, result):
+        """Extract postgres results which are just wrapped with an extra
+        list"""
         return request, map(operator.itemgetter(0), result)
 
     def tag_extract_result(self, request, result):
+        """For a tag query, we want to return a nested dict so we pipe the
+        result through this filter instead.
+        """
         rv = {}
         for uid, tn, tv in result:
             if not uid in rv: rv[uid] = {}
@@ -308,7 +313,59 @@ class Api(resource.Resource):
         else:
             return defer.succeed((request, []))
 
+    def write_one_stream(self, request, stream, stags, mime_header=False):
+        """For a CSV downlod, add some hweaders and write the data to the stream
+        """
+        writer = csv.writer(request)
+        if 'tags' in request.args and not 'none' in request.args['tags']:
+            request.write("# uuid: %s\n" % stream['uuid'])
+            request.write("# DownloadTime: " + time.ctime() + "\n")
+            request.write("# ")
+            request.write('\n# '.join((': '.join(x) for x in stags)))
+            request.write('\n')
+        map(writer.writerow, stream['Readings'])
+
+    def send_csv_reply(self, request, result, tags):
+        """CSV replies are easy"""
+        request.setHeader('Content-disposition', 'attachment; filename=%s.csv' % result[0]['uuid'])
+        self.write_one_stream(request, 
+                              result[0], 
+                              sorted(map(operator.itemgetter(1,2), tags[0][1])))
+        
+        request.finish()
+
+    def send_data_reply(self, (request, result)):
+        """After reading back some data, format it and send it to the client
+        """
+        if not 'format' in request.args or 'json' in  request.args['format']:
+            request.setHeader('Content-type', 'application/json')
+            request.write(util.json_encode(result))
+            request.finish()
+        elif 'format' in request.args and 'csv' in request.args['format']:
+            if len(result) > 1:
+                request.setResponseCode(400)
+                request.write("CSV only supported for one data stream")
+                request.finish()
+                return
+            # return cvs
+            request.setHeader('Content-type', 'text/csv')
+            if 'tags' in request.args:
+                dl = []
+                for str in result:
+                    dl.append(build_tag_query(self.db, request, [('uuid', str['uuid'])]))
+                d = defer.DeferredList(dl)
+                d.addCallback(lambda x: self.send_csv_reply(request, result, x))
+                return d
+            else:
+                return self.send_csv_reply(request, result, [(False, [])] * len(result))
+        else:
+            request.setResponseCode(400)
+            request.finish()
+
     def send_reply(self, (request, result)):
+        """Send a generic json reply.
+        """
+        request.setHeader('Content-type', 'application/json')
         request.write(util.json_encode(result))
         request.finish()
 
@@ -322,14 +379,35 @@ class Api(resource.Resource):
             return self
 
     def render_POST(self, request):
+        """The POST method is only used for sql-like queries.
+
+        The logic for parsing the query, building the true SQL
+        statement, and parsing out the results are in the queryparse
+        and querygen modules.
+        """
+        # make a parser and parse the request
         parser = qp.QueryParser(request)
-        ext, query = parser.parse(request.content.read())
-        d = self.db.runQuery(query)
-        d.addCallback(lambda x: (request, ext(x)))
+        query = request.content.read()
+        ext, query, datagetter = parser.parse(query)
+
+        # pipe the results through whatever filter is necessary
+        if not ext and not datagetter:
+            d = self.db.runOperation(query)
+        else:
+            d = self.db.runQuery(query)
+        if ext:
+            d.addCallback(lambda x: (request, ext(x)))
+        else:
+            d.addCallback(lambda x: (request, []))
+
         d.addCallback(self.send_reply)
         return server.NOT_DONE_YET
 
     def render_GET(self, request):
+        """The GET method exposes a RESTful API to ARD functions.
+
+        This lets clients look at tags and get data.
+        """
         if len(request.prepath) == 1:
             return json.dumps({'Contents': ['streams', 'query', 'data', 
                                             'next', 'prev', 'tags']})
@@ -338,13 +416,13 @@ class Api(resource.Resource):
         path = map(urllib.unquote, path)
         method = request.prepath[1]
 
+        # dispatch based on the method name
         if method != 'query':
             if len(path) % 2 != 0:
                 request.setResponseCode(400)
                 request.finish()
                 return server.NOT_DONE_YET
             path.append('uuid')            
-
         if method == 'query':
             # this allows a user to enumerate tags
             d = build_query(self.db,
@@ -352,6 +430,7 @@ class Api(resource.Resource):
                             zip(path[::2], 
                                 path[1::2] + [None]))
             d.addCallback(lambda r: self.generic_extract_result(request, r))
+            d.addCallback(self.send_reply)
         elif method == 'tags':
             # retrieve tags
             d = build_tag_query(self.db,
@@ -359,6 +438,7 @@ class Api(resource.Resource):
                                 zip(path[::2], 
                                     path[1::2] + [None]))
             d.addCallback(lambda r: self.tag_extract_result(request, r))
+            d.addCallback(self.send_reply)
         elif method in ['data', 'next', 'prev']:
             # retrieve data
             d = self.db.runQuery("""SELECT uuid, id FROM stream WHERE
@@ -366,30 +446,11 @@ id IN """ + build_inner_query(request,
                               zip(path[::2], 
                                   path[1::2] + [None]))[0])
             d.addCallback(lambda r: self.data_load_result(request, method, r))
+            d.addCallback(self.send_data_reply)
         else:
             request.setResponseCode(404)
             request.finish()
             return server.NOT_DONE_YET
 
-        d.addCallback(self.send_reply)
         return server.NOT_DONE_YET
 
-
-if __name__ == '__main__':
-    import settings
-    from twisted.enterprise import adbapi
-    from twisted.internet import reactor
-
-    # connect to the mysql db
-    cp = adbapi.ConnectionPool('MySQLdb', 
-                               host=settings.MYSQL_HOST,
-                               db=settings.MYSQL_DB,
-                               user=settings.MYSQL_USER,
-                               passwd=settings.MYSQL_PASS)
-    tr = QueryResource(cp)
-
-    d = tr.build_tree([('Metadata/Location/Campus', 'UCB'), 
-                       ('Metadata/Location/Building', 'Soda Hall'),
-                       ('Path', None)])
-    d.addCallback(lambda _:reactor.stop())
-    reactor.run()
